/*
 Honey, a UCI chess playing engine derived from Stockfish and Glaurung 2.1
 Copyright (C) 2004-2008 Tord Romstad (Glaurung author)
 Copyright (C) 2008-2015 Marco Costalba, Joona Kiiski, Tord Romstad (Stockfish Authors)
 Copyright (C) 2015-2016 Marco Costalba, Joona Kiiski, Gary Linscott, Tord Romstad (Stockfish Authors)
 Copyright (C) 2017-2020 Michael Byrne, Marco Costalba, Joona Kiiski, Gary Linscott, Tord Romstad (Honey Authors)

 Honey is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Honey is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <algorithm>
#include <cassert>
#include <cmath>
#include <cstring>   // For std::memset
#include <iostream>
#include <sstream>

#ifdef Add_Features
#include <fstream>
#include <unistd.h> //for sleep //MichaelB7
#endif

#include "evaluate.h"
#include "misc.h"
#include "movegen.h"
#include "movepick.h"
#include "position.h"
#include "search.h"
#include "thread.h"
#include "timeman.h"
#include "tt.h"
#include "uci.h"
#include "syzygy/tbprobe.h"

#ifdef Add_Features
#include "polybook.h" // Cerebellum
#endif
namespace Search {

  LimitsType Limits;
  bool adaptive;
  bool ctempt;
  int defensive;
  int dct;
  int benchKnps;
}

namespace Tablebases {

  int Cardinality;
  bool RootInTB;
  bool UseRule50;
  Depth ProbeDepth;
#ifdef Add_Features
  bool SevenManProbe; //MichaelB7
#endif
}

namespace TB = Tablebases;

using std::string;
using Eval::evaluate;
using namespace Search;

namespace {

  // Different node types, used as a template parameter
  enum NodeType { NonPV, PV };

  constexpr uint64_t ttHitAverageWindow     = 4096;
  constexpr uint64_t ttHitAverageResolution = 1024;

  // Razor and futility margins
  constexpr int RazorMargin = 531;
  Value futility_margin(Depth d, bool improving) {
    return Value(217 * (d - improving));
  }

  // Reductions lookup table, initialized at startup
  int Reductions[MAX_MOVES]; // [depth or moveNumber]
  Depth reduction(bool i, Depth d, int mn) {
    int r = Reductions[d] * Reductions[mn];
    return (r + 511) / 1024 + (!i && r > 1007);
  }

  constexpr int futility_move_count(bool improving, Depth depth) {
    return (5 + depth * depth) * (1 + improving) / 2 - 1;
  }

  // History and stats update bonus, based on depth
  int stat_bonus(Depth d) {
    return d > 15 ? -8 : 19 * d * d + 155 * d - 132;
  }

  // Add a small random component to draw evaluations to avoid 3fold-blindness
#ifndef Noir
#if defined (Sullivan)
  Value value_draw(Depth depth, Thread* thisThread) {
    return depth < 4 ? VALUE_DRAW
                   : VALUE_DRAW + Value(2 * (thisThread->nodes & 1) - 1);
  }
#else
  Value value_draw(Thread* thisThread) {
    return VALUE_DRAW + Value(2 * (thisThread->nodes & 1) - 1);
  }
#endif
#endif

  // Skill structure is used to implement strength limit
  struct Skill {
    explicit Skill(int l) : level(l) {}
    bool enabled() const { return level < 40; }
    bool time_to_pick(Depth depth) const { return depth == 1 + level; }
    Move pick_best(size_t multiPV);

    int level;
    Move best = MOVE_NONE;
  };
#ifdef Weakfish
bool  weakFishSearch;
#endif
#ifdef Add_Features
bool  fide, jekyll, minOutput, uci_sleep;
bool limitStrength = false;
int  intLevel = 40, tactical, uci_elo;
#endif

  // Breadcrumbs are used to mark nodes as being searched by a given thread
  struct Breadcrumb {
    std::atomic<Thread*> thread;
    std::atomic<Key> key;
  };
  std::array<Breadcrumb, 1024> breadcrumbs;

  // ThreadHolding structure keeps track of which thread left breadcrumbs at the given
  // node for potential reductions. A free node will be marked upon entering the moves
  // loop by the constructor, and unmarked upon leaving that loop by the destructor.
  struct ThreadHolding {
    explicit ThreadHolding(Thread* thisThread, Key posKey, int ply) {
       location = ply < 8 ? &breadcrumbs[posKey & (breadcrumbs.size() - 1)] : nullptr;
       otherThread = false;
       owning = false;
       if (location)
       {
          // See if another already marked this location, if not, mark it ourselves
          Thread* tmp = (*location).thread.load(std::memory_order_relaxed);
          if (tmp == nullptr)
          {
              (*location).thread.store(thisThread, std::memory_order_relaxed);
              (*location).key.store(posKey, std::memory_order_relaxed);
              owning = true;
          }
          else if (   tmp != thisThread
                   && (*location).key.load(std::memory_order_relaxed) == posKey)
              otherThread = true;
       }
    }

    ~ThreadHolding() {
       if (owning) // Free the marked location
           (*location).thread.store(nullptr, std::memory_order_relaxed);
    }

    bool marked() { return otherThread; }

    private:
    Breadcrumb* location;
    bool otherThread, owning;
  };

  template <NodeType NT>
  Value search(Position& pos, Stack* ss, Value alpha, Value beta, Depth depth, bool cutNode);

  template <NodeType NT>
  Value qsearch(Position& pos, Stack* ss, Value alpha, Value beta, Depth depth = 0);

  Value value_to_tt(Value v, int ply);
  Value value_from_tt(Value v, int ply, int r50c);
  void update_pv(Move* pv, Move move, Move* childPv);
  void update_continuation_histories(Stack* ss, Piece pc, Square to, int bonus);
  void update_quiet_stats(const Position& pos, Stack* ss, Move move, int bonus);
  void update_all_stats(const Position& pos, Stack* ss, Move bestMove, Value bestValue, Value beta, Square prevSq,
                        Move* quietsSearched, int quietCount, Move* capturesSearched, int captureCount, Depth depth);

  // perft() is our utility to verify move generation. All the leaf nodes up
  // to the given depth are generated and counted, and the sum is returned.
  template<bool Root>
  uint64_t perft(Position& pos, Depth depth) {

    StateInfo st;
    uint64_t cnt, nodes = 0;
    const bool leaf = (depth == 2);

    for (const auto& m : MoveList<LEGAL>(pos))
    {
        if (Root && depth <= 1)
            cnt = 1, nodes++;
        else
        {
            pos.do_move(m, st);
            cnt = leaf ? MoveList<LEGAL>(pos).size() : perft<false>(pos, depth - 1);
            nodes += cnt;
            pos.undo_move(m);
        }
        if (Root)
            sync_cout << UCI::move(m, pos.is_chess960()) << ": " << cnt << sync_endl;
    }
    return nodes;
  }

} // namespace


/// Search::init() is called at startup to initialize various lookup tables

void Search::init() {

  for (int i = 1; i < MAX_MOVES; ++i)
      Reductions[i] = int((24.8 + std::log(Threads.size()) / 2) * std::log(i));
}
/// Search::clear() resets search state to its initial value

void Search::clear() {

  Threads.main()->wait_for_search_finished();

  Time.availableNodes = 0;
  TT.clear();
  Threads.clear();
  Tablebases::init(Options["SyzygyPath"]); // Free mapped files
}


/// MainThread::search() is started when the program receives the UCI 'go'
/// command. It searches from the root position and outputs the "bestmove".

void MainThread::search() {

  if (Limits.perft)
  {
      nodes = perft<true>(rootPos, Limits.perft);
      sync_cout << "\nNodes searched: " << nodes << "\n" << sync_endl;
      return;
  }
#ifdef Add_Features
    PRNG rng(now());
    int shallowBlue_adjust = 35; //to roughly anchor 1712 rating to CCRL Shallow BLue 2.0 rating of 1712
	/* Reset on 11/14/2019:
  500 game(s) loaded
  Rank Name                    Rating   Δ     +    -     #     Σ    Σ%     W    L    D   W%    =%   OppR
  ---------------------------------------------------------------------------------------------------------
  1 Honey-XR7-1712  11/14/2019  1055   0.0   25   25   500  251.5  50.3  169  166  165  33.8  33.0  1053
  2 Honey-CCRL-1712 10/02/2019  1053   1.3   25   25   500  248.5  49.7  166  169  165  33.2  33.0  1055
  ---------------------------------------------------------------------------------------------------------*/
    adaptive            = Options["Adaptive_Play"];
#ifdef Weakfish
    weakFishSearch      = Options["WeakFish"];
#endif
    defensive           = Options["Defensive"];
    fide                = Options["FIDE_Ratings"];
    jekyll              = Options["Variety"];
    minOutput           = Options["Minimal_Output"];
    tactical            = Options["Tactical"];
    uci_elo             = Options["Engine_Elo"];
    uci_sleep           = Options["UCI_Sleep"];


#endif

  Color us = rootPos.side_to_move();
  Time.init(Limits, us, rootPos.game_ply());
  TT.new_search();

  if (rootMoves.empty())
  {
      rootMoves.emplace_back(MOVE_NONE);
      sync_cout << "info depth 0 score "
                << UCI::value(rootPos.checkers() ? -VALUE_MATE : VALUE_DRAW)
                << sync_endl;
  }
  else
  {
#ifdef Add_Features
      Move bookMove = MOVE_NONE;

      if (bool(Options["Use_Book_1"]) && !Limits.infinite && !Limits.mate)
          bookMove = polybook1.probe(rootPos);
      if (bool(Options["Use_Book_2"]) && !bookMove && !Limits.infinite && !Limits.mate)
          bookMove = polybook2.probe(rootPos);
      if (bool(Options["Use_Book_3"]) && !bookMove && !Limits.infinite && !Limits.mate)
          bookMove = polybook3.probe(rootPos);
      if (bool(Options["Use_Book_4"]) && !bookMove && !Limits.infinite && !Limits.mate)
          bookMove = polybook4.probe(rootPos);

      if (bookMove && std::count(rootMoves.begin(), rootMoves.end(), bookMove))
      {
          for (Thread* th : Threads)
              std::swap(th->rootMoves[0], *std::find(th->rootMoves.begin(), th->rootMoves.end(), bookMove));
      }
      else
      {
         if (Options["NPS_Level"])
         {
             benchKnps = 1000 * (Options["Bench_KNPS"]);
             sync_cout << "Bknps: " << benchKnps  << sync_endl;// for debug
             int nps = 64 * pow(1.1486, (Options["NPS_Level"] - 1 ));
             Limits.nodes = nps;
             Limits.nodes *= Time.optimum()/1000 + 1 ;
             std::this_thread::sleep_for (std::chrono::milliseconds(Time.optimum()) * double(1 - Limits.nodes/benchKnps));
         }
         else if (Options["UCI_LimitStrength"] && Options["Engine_Level"] == "None")
         {
             uci_elo = (Options["UCI_Elo"]);
             limitStrength = true;
			 jekyll=true; //on with uci_elo, variety gets turned off with adaptive
             goto skipLevels;
         }
         if (Options["Engine_Level"] == "None")
         {
             limitStrength = false;
             goto skipLevels;
         }
         else
             {
				 limitStrength = true;
				 jekyll = true;
			 }

         if (Options["Engine_Level"] == "World_Champion")
             uci_elo = 2900;
         else if (Options["Engine_Level"] == "Super_GM")
             uci_elo = 2800;
         else if (Options["Engine_Level"] == "GM")
             uci_elo = 2650;
         else if (Options["Engine_Level"] == "Deep_Thought")
             uci_elo = 2500;
         else if (Options["Engine_Level"] == "SIM")
             uci_elo = 2400;
		 else if (Options["Engine_Level"] == "IM")
             uci_elo = 2300;
         else if (Options["Engine_Level"] == "Cray_Blitz")
             uci_elo = 2200;
         else if (Options["Engine_Level"] == "Master")
             uci_elo = 2100;
         else if (Options["Engine_Level"] == "Expert")
             uci_elo = 1950;
         else if (Options["Engine_Level"] == "Class_A")
             uci_elo = 1800;
         else if (Options["Engine_Level"] == "Class_B")
             uci_elo = 1650;
         else if (Options["Engine_Level"] == "Class_C")
             uci_elo = 1500;
         else if (Options["Engine_Level"] == "Class_D")
             uci_elo = 1350;
         else if (Options["Engine_Level"] == "Boris")
             uci_elo = 1200;
         else if (Options["Engine_Level"] == "Novice")
             uci_elo = 1000;
skipLevels:
         if (limitStrength)
         {  //note varietry strength is capped around ~2150-2200 due to its robustness
             benchKnps = 1000 * (Options["Bench_KNPS"]);
             int random = (rand() % 20 - 10);
             uci_elo = uci_elo + random + shallowBlue_adjust;
             //sync_cout << "Elo " << uci_elo << sync_endl;// for debug
             int ccrlELo = uci_elo;
             if (fide)
                 uci_elo = (((uci_elo * 10) / 7) - 1200);  //shallowBlue adj was only required to get CCRL rating correct
             uci_elo += 200; //  to offset Elo loss with variety
             uci_elo = std::min(uci_elo, 3200);

             int NodesToSearch  =  pow(1.00382, (uci_elo - 999)) * 48;
             //sync_cout << "Nodes To Search: " << NodesToSearch << sync_endl;//for debug
             Limits.nodes = NodesToSearch;
             Limits.nodes *= Time.optimum()/1000;
             Limits.nodes = clamp(Limits.nodes, (int64_t) 48,Limits.nodes);
             //int sleepValue = Time.optimum() * double(1 - Limits.nodes/benchKnps);
             //sync_cout << "Sleep time: " << sleepValue << sync_endl;//for debug
             //sync_cout << "Limit Nodes: " <<  Limits.nodes << sync_endl;//for debug
             if (uci_sleep)
                 std::this_thread::sleep_for (std::chrono::milliseconds(Time.optimum()) * double(1 - Limits.nodes/benchKnps));
             uci_elo =  ccrlELo - shallowBlue_adjust;
         }
#endif
      for (Thread* th : Threads)
      {
          th->bestMoveChanges = 0;
          if (th != this)
              th->start_searching();
      }

      Thread::search(); // Let's start searching!
#ifdef Add_Features
    }
#endif
  }

  // When we reach the maximum depth, we can arrive here without a raise of
  // Threads.stop. However, if we are pondering or in an infinite search,
  // the UCI protocol states that we shouldn't print the best move before the
  // GUI sends a "stop" or "ponderhit" command. We therefore simply wait here
  // until the GUI sends one of those commands.

  while (!Threads.stop && (ponder || Limits.infinite))
  {} // Busy wait for a stop or a ponder reset

  // Stop the threads if not already stopped (also raise the stop if
  // "ponderhit" just reset Threads.ponder).
  Threads.stop = true;

  // Wait until all threads have finished
  for (Thread* th : Threads)
      if (th != this)
          th->wait_for_search_finished();

  // When playing in 'nodes as time' mode, subtract the searched nodes from
  // the available ones before exiting.
  if (Limits.npmsec)
      Time.availableNodes += Limits.inc[us] - Threads.nodes_searched();

  Thread* bestThread = this;

  // Check if there are threads with a better score than main thread
#if defined (Sullivan) || (Noir)  //joergoster 3240f204 - check all threads of rmate -in -x
  if (   (Options["MultiPV"] == 1 || Limits.mate)
#else
  if (    Options["MultiPV"] == 1
#endif
      && !Limits.depth
#ifdef Add_Features
      && !(Skill(Options["Skill Level"]).enabled() || Options["UCI_LimitStrength"] || limitStrength)
#else
      && !(Skill(Options["Skill Level"]).enabled())
#endif
      &&  rootMoves[0].pv[0] != MOVE_NONE)
  {
      std::map<Move, int64_t> votes;
      Value minScore = this->rootMoves[0].score;

      // Find out minimum score
      for (Thread* th: Threads)
          minScore = std::min(minScore, th->rootMoves[0].score);

      // Vote according to score and depth, and select the best thread
      for (Thread* th : Threads)
      {
          votes[th->rootMoves[0].pv[0]] +=
              (th->rootMoves[0].score - minScore + 14) * int(th->completedDepth);

          if (bestThread->rootMoves[0].score >= VALUE_MATE_IN_MAX_PLY)
          {
              // Make sure we pick the shortest mate
              if (th->rootMoves[0].score > bestThread->rootMoves[0].score)
                  bestThread = th;
          }
          else if (   th->rootMoves[0].score >= VALUE_MATE_IN_MAX_PLY
                   || votes[th->rootMoves[0].pv[0]] > votes[bestThread->rootMoves[0].pv[0]])
              bestThread = th;
      }
  }
      previousScore = bestThread->rootMoves[0].score;

  // Send again PV info if we have a new best thread
  if (bestThread != this || Skill(Options["Skill Level"]).enabled())
      sync_cout << UCI::pv(bestThread->rootPos, bestThread->completedDepth, -VALUE_INFINITE, VALUE_INFINITE) << sync_endl;

  if  (adaptive ) //mutually exclusive with variety
  {
#ifdef Add_Features
	  jekyll = false;
#endif
      size_t i = 0;
      if ( previousScore >= -PawnValueMg && previousScore <= PawnValueMg * 4 )
	  {
          while (i+1 < rootMoves.size() && bestThread->rootMoves[i+1].score > previousScore)
          ++i;
          previousScore = bestThread->rootMoves[i].score;
          sync_cout << UCI::pv(bestThread->rootPos, bestThread->completedDepth, -VALUE_INFINITE, VALUE_INFINITE) << sync_endl;
          sync_cout << "bestmove " << UCI::move(bestThread->rootMoves[i].pv[0], rootPos.is_chess960());
	  }
      else if ( previousScore > PawnValueMg * 4  && previousScore <  PawnValueMg * 7 )
      {
          while (i+1 < rootMoves.size() && bestThread->rootMoves[i+1].score < previousScore
                && previousScore + PawnValueMg/2  > bestThread->rootMoves[i+1].score)
		  {
              ++i;
              break;
          }
          previousScore = bestThread->rootMoves[i].score;
          while (i+1 < rootMoves.size() && bestThread->rootMoves[i+1].score > previousScore
                && previousScore + PawnValueMg/2  < bestThread->rootMoves[i+1].score)
          {
              ++i;
              previousScore = bestThread->rootMoves[i-1].score;

          }
          previousScore = bestThread->rootMoves[i].score;
          sync_cout << UCI::pv(bestThread->rootPos, bestThread->completedDepth, -VALUE_INFINITE, VALUE_INFINITE) << sync_endl;
          sync_cout << "bestmove " << UCI::move(bestThread->rootMoves[i].pv[0], rootPos.is_chess960());
      }
      else
      {
          sync_cout << "bestmove " << UCI::move(bestThread->rootMoves[0].pv[0], rootPos.is_chess960());
          if (bestThread->rootMoves[0].pv.size() > 1 || bestThread->rootMoves[0].extract_ponder_from_tt(rootPos))
              std::cout << " ponder " << UCI::move(bestThread->rootMoves[0].pv[1], rootPos.is_chess960());
      }
  }
  else
  {
  sync_cout << "bestmove " << UCI::move(bestThread->rootMoves[0].pv[0], rootPos.is_chess960());

  if (bestThread->rootMoves[0].pv.size() > 1 || bestThread->rootMoves[0].extract_ponder_from_tt(rootPos))
      std::cout << " ponder " << UCI::move(bestThread->rootMoves[0].pv[1], rootPos.is_chess960());
  }
  std::cout << sync_endl;
}


/// Thread::search() is the main iterative deepening loop. It calls search()
/// repeatedly with increasing depth until the allocated thinking time has been
/// consumed, the user stops the search, or the maximum search depth is reached.

void Thread::search() {

  // To allow access to (ss-7) up to (ss+2), the stack must be oversized.
  // The former is needed to allow update_continuation_histories(ss-1, ...),
  // which accesses its argument at ss-6, also near the root.
  // The latter is needed for statScores and killer initialization.
  Stack stack[MAX_PLY+10], *ss = stack+7;
  Move  pv[MAX_PLY+1];
  Value bestValue, alpha, beta, delta;
  Move  lastBestMove = MOVE_NONE;
  Depth lastBestMoveDepth = 0;
  MainThread* mainThread = (this == Threads.main() ? Threads.main() : nullptr);
  double timeReduction = 1, totBestMoveChanges = 0;
  Color us = rootPos.side_to_move();
  ctempt= Options["Contempt"];
  defensive = -34 * (int(Options["Defensive"]));  //about 16 cp

#ifdef Add_Features
  TB::SevenManProbe = Options["7 Man Probing"];
#endif
#if defined (Stockfish) || (Weakfish)
  int iterIdx = 0;
#endif
  std::memset(ss-7, 0, 10 * sizeof(Stack));
  for (int i = 7; i > 0; i--)

      (ss-i)->continuationHistory = &this->continuationHistory[0][0][NO_PIECE][0]; // Use as a sentinel

  ss->pv = pv;

  bestValue = delta = alpha = -VALUE_INFINITE;
  beta = VALUE_INFINITE;
#if defined (Stockfish) || (Weakfish)
  if (mainThread)
  {
      if (mainThread->previousScore == VALUE_INFINITE)
          for (int i=0; i<4; ++i)
              mainThread->iterValue[i] = VALUE_ZERO;
      else
          for (int i=0; i<4; ++i)
              mainThread->iterValue[i] = mainThread->previousScore;
  }
#endif
  size_t multiPV = Options["MultiPV"];
  PRNG rng(now());
#ifndef Add_Features
  // Pick integer skill levels, but non-deterministically round up or down
  // such that the average integer skill corresponds to the input floating point one.
  // UCI_Elo is converted to a suitable fractional skill level, using anchoring
  // to CCRL Elo (goldfish 1.13 = 2000) and a fit through Ordo derived Elo
  // for match (TC 60+0.6) results spanning a wide range of k values.

  double floatLevel = Options["UCI_LimitStrength"] ?
                        clamp(std::pow((Options["UCI_Elo"] - 1346.6) / 71.7, 1 / 0.806), 0.0, 40.0) :
                        double(Options["Skill Level"]);
  int intLevel = int(floatLevel) +
                 ((floatLevel - int(floatLevel)) * 1024 > rng.rand<unsigned>() % 1024  ? 1 : 0);
#endif
	Skill skill(intLevel);

#ifdef Add_Features
    if (tactical) multiPV = pow(2, tactical);
#endif

  // When playing with strength handicap enable MultiPV search that we will
  // use behind the scenes to retrieve a set of possible moves.
  if (skill.enabled())
      multiPV = std::max(multiPV, (size_t)4);

  multiPV = std::min(multiPV, rootMoves.size());
  ttHitAverage = ttHitAverageWindow * ttHitAverageResolution / 2;

int ct = int(ctempt) * (int(Options["Contempt_Value"]) * PawnValueEg / 100); // From centipawns
  // In analysis mode, adjust contempt in accordance with user preference
  if (Limits.infinite || Options["UCI_AnalyseMode"])
      ct =  Options["Analysis_Contempt"] == "Off"  ? 0
          : Options["Analysis_Contempt"] == "Both" ? ct
          : Options["Analysis_Contempt"] == "White" && us == BLACK ? -ct
          : Options["Analysis_Contempt"] == "Black" && us == WHITE ? -ct
          : ct;

  // Evaluation score is from the white point of view
  contempt = (us == WHITE ?  make_score(ct, ct / 2)
                          : -make_score(ct, ct / 2));

  int searchAgainCounter = 0;

  // Iterative deepening loop until requested to stop or the target depth is reached
  while (   ++rootDepth < MAX_PLY
         && !Threads.stop
         && !(Limits.depth && mainThread && rootDepth > Limits.depth))
  {
      // Age out PV variability metric
      if (mainThread)
          totBestMoveChanges /= 2;

      // Save the last iteration's scores before first PV line is searched and
      // all the move scores except the (new) PV are set to -VALUE_INFINITE.
      for (RootMove& rm : rootMoves)
          rm.previousScore = rm.score;

      size_t pvFirst = 0;
      pvLast = 0;

      if (!Threads.increaseDepth)
         searchAgainCounter++;

      // MultiPV loop. We perform a full root search for each PV line
      for (pvIdx = 0; pvIdx < multiPV && !Threads.stop; ++pvIdx)
      {
          if (pvIdx == pvLast)
          {
              pvFirst = pvLast;
              for (pvLast++; pvLast < rootMoves.size(); pvLast++)
                  if (rootMoves[pvLast].tbRank != rootMoves[pvFirst].tbRank)
                      break;
          }

          // Reset UCI info selDepth for each depth and each PV line
          selDepth = 0;

          // Reset aspiration window starting size
          if (rootDepth >= 4)
          {
              Value previousScore = rootMoves[pvIdx].previousScore;

#if defined (Sullivan) || (Blau) || (Fortress) || (Noir)
              delta = Value(20 + abs(previousScore) / 64);
#else
              delta = Value(21 + abs(previousScore) / 256);
#endif
              alpha = std::max(previousScore - delta,-VALUE_INFINITE);
              beta  = std::min(previousScore + delta, VALUE_INFINITE);
#if defined (Sullivan) || (Blau) || (Fortress) || (Noir)
			  // Adjust contempt based on root move's previousScore (dynamic contempt)
              dct = ctempt * (ct + (111 - ct / 2) * previousScore / (abs(previousScore) + 176)) + defensive;
#else
              dct = ctempt * (ct + (102 - ct / 2) * previousScore / (abs(previousScore) + 157)) + defensive;
#endif
              contempt = (us == WHITE ?  make_score(dct, dct / 2)
                       : -make_score(dct, dct / 2));
       }

          // Start with a small aspiration window and, in the case of a fail
          // high/low, re-search with a bigger window until we don't fail
          // high/low anymore.
#ifndef Noir
          int failedHighCnt = 0;
#endif
          while (true )
          {
#ifdef Noir
              bestValue = ::search<PV>(rootPos, ss, alpha, beta, rootDepth, false);
#else
              Depth adjustedDepth = std::max(1, rootDepth - failedHighCnt - searchAgainCounter);
              bestValue = ::search<PV>(rootPos, ss, alpha, beta, adjustedDepth, false);
#endif

              // Bring the best move to the front. It is critical that sorting
              // is done with a stable algorithm because all the values but the
              // first and eventually the new best one are set to -VALUE_INFINITE
              // and we want to keep the same order for all the moves except the
              // new PV that goes to the front. Note that in case of MultiPV
              // search the already searched PV lines are preserved.
              std::stable_sort(rootMoves.begin() + pvIdx, rootMoves.begin() + pvLast);

              // If search has been stopped, we break immediately. Sorting is
              // safe because RootMoves is still valid, although it refers to
              // the previous iteration.
              if (Threads.stop)
                  break;

                // When failing high/low give some update (without cluttering
                // the UI) before a re-search.
                if (
#ifdef Add_Features
                    !minOutput &&
#endif
                    mainThread
                    && multiPV == 1
                    && (bestValue <= alpha || bestValue >= beta)
                    && Time.elapsed() > 3000)
                    sync_cout << UCI::pv(rootPos, rootDepth, alpha, beta) << sync_endl;

              // In case of failing low/high increase aspiration window and
              // re-search, otherwise exit the loop.

              if (bestValue <= alpha)
              {
                  beta = (alpha + beta) / 2;
                  alpha = std::max(bestValue - delta, -VALUE_INFINITE);

#ifndef Noir
                  failedHighCnt = 0;
#endif
                  if (mainThread)
                      mainThread->stopOnPonderhit = false;
              }
              else if (bestValue >= beta)
              {
                  beta = std::min(bestValue + delta, VALUE_INFINITE);
#ifndef Noir
                  ++failedHighCnt;
#endif
              }
              else
			  {
                  ++rootMoves[pvIdx].bestMoveCount;
				  break;
              }
              delta += delta / 4 + 5;

              assert(alpha >= -VALUE_INFINITE && beta <= VALUE_INFINITE);
          }

          // Sort the PV lines searched so far and update the GUI
          std::stable_sort(rootMoves.begin() + pvFirst, rootMoves.begin() + pvIdx + 1);

          if (    mainThread
              && (Threads.stop || pvIdx + 1 == multiPV || Time.elapsed() > 3000))
              sync_cout << UCI::pv(rootPos, rootDepth, alpha, beta) << sync_endl;
              //int dctcp = dct * 100/PawnValueEg; //for debug
              //sync_cout << "Contempt MG: " << dctcp  << sync_endl;// for debug
              //sync_cout << "Contempt EG: " << dctcp/2  << "\n" << sync_endl;// for debug
      }

        if (!Threads.stop)
          completedDepth = rootDepth;
      if (rootMoves[0].pv[0] != lastBestMove) {
         lastBestMove = rootMoves[0].pv[0];
         lastBestMoveDepth = rootDepth;
      }

      // Have we found a "mate in x"?
      if (   Limits.mate
             && bestValue >= VALUE_MATE_IN_MAX_PLY
             && VALUE_MATE - bestValue <= 2 * Limits.mate)
         Threads.stop = true;

	  if (!mainThread)
		  continue;

#ifdef Add_Features
        if (Options["FastPlay"])
        {
            if ( Time.elapsed() > Time.optimum() / 256
                 && ( abs(bestValue) > 12300 ||  abs(bestValue) >= VALUE_MATE_IN_MAX_PLY ))
                 Threads.stop = true;
        }
#endif
      // If skill level is enabled and time is up, pick a sub-optimal best move
      if (skill.enabled() && skill.time_to_pick(rootDepth))
          skill.pick_best(multiPV);

      // Do we have time for the next iteration? Can we stop searching now?
      if (    Limits.use_time_management()
          && !Threads.stop
          && !mainThread->stopOnPonderhit)
      {
#if defined (Stockfish) || (Weakfish)
          double fallingEval = (332 +  6 * (mainThread->previousScore - bestValue)
                                    +  6 * (mainThread->iterValue[iterIdx]  - bestValue)) / 704.0;
#else
          double fallingEval = (354 + 10 * (mainThread->previousScore - bestValue)) / 692.0;
#endif
          fallingEval = clamp(fallingEval, 0.5, 1.5);

          // If the bestMove is stable over several iterations, reduce time accordingly
          timeReduction = lastBestMoveDepth + 9 < completedDepth ? 1.94 : 0.91;
          double reduction = (1.41 + mainThread->previousTimeReduction) / (2.27 * timeReduction);

          // Use part of the gained time from a previous stable move for the current move
          for (Thread* th : Threads)
          {
              totBestMoveChanges += th->bestMoveChanges;
              th->bestMoveChanges = 0;
          }
          double bestMoveInstability = 1 + totBestMoveChanges / Threads.size();

          // Stop the search if we have only one legal move, or if available time elapsed
          if (   rootMoves.size() == 1
              || Time.elapsed() > Time.optimum() * fallingEval * reduction * bestMoveInstability)
          {
              // If we are allowed to ponder do not stop the search now but
              // keep pondering until the GUI sends "ponderhit" or "stop".
              if (mainThread->ponder)
                  mainThread->stopOnPonderhit = true;
              else
                  Threads.stop = true;
          }
          else if (   Threads.increaseDepth
                   && !mainThread->ponder
                   && Time.elapsed() > Time.optimum() * fallingEval * reduction * bestMoveInstability * 0.6)
                   Threads.increaseDepth = false;
          else
                   Threads.increaseDepth = true;
      }
#if defined (Stockfish) || (Weakfish)
      mainThread->iterValue[iterIdx] = bestValue;
      iterIdx = (iterIdx + 1) & 3;
#endif
  }

  if (!mainThread)
      return;

  mainThread->previousTimeReduction = timeReduction;

  // If skill level is enabled, swap best PV line with the sub-optimal one
  if (skill.enabled())
      std::swap(rootMoves[0], *std::find(rootMoves.begin(), rootMoves.end(),
                skill.best ? skill.best : skill.pick_best(multiPV)));
}


namespace {
#if defined (Sullivan) || (Blau) || (Fortress)
  static TTEntry *probeTT(const Position &pos, Stack* ss, Key posKey,
                          bool &ttHit, Value &ttValue, Move &ttMove) {

    TTEntry *tte = TT.probe(posKey, ttHit);
    ttValue = tte->value();
    // Disregard TT hit if mate score is shorter than remaining 50 MR plies.
    if (   ttHit
        && ttValue != VALUE_NONE
        && std::abs(ttValue) >= VALUE_MATE_IN_MAX_PLY
        && VALUE_MATE - std::abs(ttValue) >= 100 - pos.rule50_count())
        ttHit = false;
    if (   ttHit
        && ttValue != VALUE_NONE
		&& std::abs(ttValue) >= VALUE_MATE_IN_MAX_PLY
        && pos.count<ALL_PIECES>() < 8)
        ttHit = true;
    ttValue = ttHit ? value_from_tt(tte->value(), ss->ply, pos.rule50_count()) : VALUE_NONE;
    ttMove = ttHit ? tte->move() : MOVE_NONE;
    return tte;
  }
#endif
  // search<>() is the main search function for both PV and non-PV nodes

  template <NodeType NT>
  Value search(Position& pos, Stack* ss, Value alpha, Value beta, Depth depth, bool cutNode) {

    constexpr bool PvNode = NT == PV;
    const bool rootNode = PvNode && ss->ply == 0;
#ifndef Fortress
#ifndef Noir
    // Check if we have an upcoming move which draws by repetition, or
    // if the opponent had an alternative move earlier to this position.
#if defined (Sullivan) || (Blau) || (Fortress) || (Noir)
    if (   pos.rule50_count() >= 5
#else
    if (   pos.rule50_count() >= 3
#endif
        && alpha < VALUE_DRAW
        && !rootNode
        && pos.has_game_cycle(ss->ply))
    {
#if defined (Sullivan) || (Noir)
        alpha = value_draw(depth, pos.this_thread());
#else
        alpha = value_draw(pos.this_thread());
#endif
        if (alpha >= beta)
            return alpha;
    }
#endif
#endif
    // Dive into quiescence search when the depth reaches zero
    if (depth <= 0)
        return qsearch<NT>(pos, ss, alpha, beta);

    assert(-VALUE_INFINITE <= alpha && alpha < beta && beta <= VALUE_INFINITE);
    assert(PvNode || (alpha == beta - 1));
    assert(0 < depth && depth < MAX_PLY);
    assert(!(PvNode && cutNode));

    Move pv[MAX_PLY+1], capturesSearched[32], quietsSearched[64];
	  StateInfo st;
    TTEntry* tte;
    Key posKey;
    Move ttMove, move, excludedMove, bestMove;
    Depth extension, newDepth;
#ifdef Noir
    Value bestValue, value, ttValue, eval;
    bool ttHit, ttPv, inCheck, givesCheck, improving, didLMR, priorCapture, isMate, gameCycle;
#elif defined Fortress
    bool ttHit, ttPv, inCheck, givesCheck, improving, didLMR, priorCapture, gameCycle;
    Value bestValue, value, ttValue, eval, maxValue;
#else
    bool ttHit, ttPv, inCheck, givesCheck, improving, didLMR, priorCapture;
    Value bestValue, value, ttValue, eval, maxValue;
#endif
    bool captureOrPromotion, doFullDepthSearch, moveCountPruning, ttCapture, singularLMR;
    Piece movedPiece;
    int moveCount, captureCount, quietCount;

    // Step 1. Initialize node
    Thread* thisThread = pos.this_thread();
    inCheck = pos.checkers();
    priorCapture = pos.captured_piece();
    Color us = pos.side_to_move();
    moveCount = captureCount = quietCount = ss->moveCount = 0;
    bestValue = -VALUE_INFINITE;
#ifndef Noir
    maxValue = VALUE_INFINITE;
#endif
#if defined (Fortress) || (Noir)
    gameCycle = false;
#endif
    // Check for the available remaining time
    if (thisThread == Threads.main())
        static_cast<MainThread*>(thisThread)->check_time();

    // Used to send selDepth info to GUI (selDepth counts from 1, ply from 0)
    if (PvNode && thisThread->selDepth < ss->ply + 1)
        thisThread->selDepth = ss->ply + 1;

#if defined (Noir) || (Fortress)
    // Transposition table lookup. We don't want the score of a partial
    // search to overwrite a previous full search TT value, so we use a different
    // position key in case of an excluded move.
    excludedMove = ss->excludedMove;
#ifdef Noir
    posKey = pos.key() ^ Key(excludedMove);
#endif
#ifdef Fortress
    posKey = pos.key() ^ Key(excludedMove) << 16;
#endif
    tte = TT.probe(posKey, ttHit);
    ttValue = ttHit ? value_from_tt(tte->value(), ss->ply, pos.rule50_count()) : VALUE_NONE;
    ttMove =  rootNode ? thisThread->rootMoves[thisThread->pvIdx].pv[0]
              : ttHit    ? tte->move() : MOVE_NONE;
    ttPv = PvNode || (ttHit && tte->is_pv());

    // thisThread->ttHitAverage can be used to approximate the running average of ttHit
    thisThread->ttHitAverage =   (ttHitAverageWindow - 1) * thisThread->ttHitAverage / ttHitAverageWindow
    + ttHitAverageResolution * ttHit;
#endif
    if (!rootNode)
    {
#if defined (Fortress) || (Noir)
        // Check if we have an upcoming move which draws by repetition, or
        // if the opponent had an alternative move earlier to this position.
        if (pos.has_game_cycle(ss->ply))
        {
            if (VALUE_DRAW >= beta)
            {
                tte->save(posKey, VALUE_DRAW, ttPv, BOUND_EXACT,
                          depth, MOVE_NONE, VALUE_NONE);

                return VALUE_DRAW;
            }
            gameCycle = true;
            alpha = std::max(alpha, VALUE_DRAW);
        }

        // Step 2. Check for aborted search and immediate draw
        if (pos.is_draw(ss->ply))
            return VALUE_DRAW;
#endif

        // Step 2. Check for aborted search and immediate draw
#if defined (Fortress) || (Noir)
        if (   Threads.stop.load(std::memory_order_relaxed) || ss->ply >= MAX_PLY)
#else
        if (   Threads.stop.load(std::memory_order_relaxed)
            || pos.is_draw(ss->ply)
            || ss->ply >= MAX_PLY)
#endif
            return ss->ply >= MAX_PLY && !inCheck ? evaluate(pos)
#if defined (Fortress) || (Noir)
                                                    : VALUE_DRAW;
#elif defined (Sullivan)
                                                    : value_draw(depth, pos.this_thread());
#else
                                                    : value_draw(pos.this_thread());
#endif
        // Step 3. Mate distance pruning. Even if we mate at the next move our score
        // would be at best mate_in(ss->ply+1), but if alpha is already bigger because
        // a shorter mate was found upward in the tree then there is no need to search
        // because we will never beat the current alpha. Same logic but with reversed
        // signs applies also in the opposite condition of being mated instead of giving
        // mate. In this case return a fail-high score.
#ifdef Noir
        if (alpha >= mate_in(ss->ply+1))
#else
        alpha = std::max(mated_in(ss->ply), alpha);
        beta = std::min(mate_in(ss->ply+1), beta);
        if (alpha >= beta)
#endif

            return alpha;
    }

    assert(0 <= ss->ply && ss->ply < MAX_PLY);

    (ss+1)->ply = ss->ply + 1;
    (ss+1)->excludedMove = bestMove = MOVE_NONE;
    (ss+2)->killers[0] = (ss+2)->killers[1] = MOVE_NONE;
    Square prevSq = to_sq((ss-1)->currentMove);

    // Initialize statScore to zero for the grandchildren of the current position.
    // So statScore is shared between all grandchildren and only the first grandchild
    // starts with statScore = 0. Later grandchildren start with the last calculated
    // statScore of the previous grandchild. This influences the reduction rules in
    // LMR which are based on the statScore of parent position.
    if (rootNode)
        (ss+4)->statScore = 0;
    else
        (ss+2)->statScore = 0;
#ifndef Noir
#ifndef Fortress

    // Step 4. Transposition table lookup. We don't want the score of a partial
    // search to overwrite a previous full search TT value, so we use a different
    // position key in case of an excluded move.
    excludedMove = ss->excludedMove;
    posKey = pos.key() ^ Key(excludedMove << 16); // Isn't a very good hash
#if defined (Sullivan) || (Blau)
    tte = probeTT(pos, ss, posKey, ttHit, ttValue, ttMove);
#else
    tte = TT.probe(posKey, ttHit);
#endif
    ttValue = ttHit ? value_from_tt(tte->value(), ss->ply, pos.rule50_count()) : VALUE_NONE;
    ttMove =  rootNode ? thisThread->rootMoves[thisThread->pvIdx].pv[0]
              : ttHit    ? tte->move() : MOVE_NONE;

    ttPv = PvNode || (ttHit && tte->is_pv());

    // thisThread->ttHitAverage can be used to approximate the running average of ttHit
    thisThread->ttHitAverage =   (ttHitAverageWindow - 1) * thisThread->ttHitAverage / ttHitAverageWindow
                                + ttHitAverageResolution * ttHit;
#endif
#endif

#ifdef Noir
    int piecesCount = popcount(pos.pieces());
#else
    int piecesCount = pos.count<ALL_PIECES>();
#endif
    // At non-PV nodes we check for an early TT cutoff
    if (  !PvNode
        && ttHit
#if defined (Sullivan) || (Blau)
    && (pos.rule50_count() < 92 || (piecesCount < 8  && TB::Cardinality ))
#endif
#if defined (Fortress) || (Noir)
        && !gameCycle
#endif
        && tte->depth() >= depth
        && ttValue != VALUE_NONE // Possible in case of TT access race
#if defined (Fortress) || (Noir)
        && (ttValue != VALUE_DRAW || VALUE_DRAW >= beta)
#endif
        && (ttValue >= beta ? (tte->bound() & BOUND_LOWER)
                            : (tte->bound() & BOUND_UPPER)))
    {
        // If ttMove is quiet, update move sorting heuristics on TT hit
        if (ttMove)
        {
            if (ttValue >= beta)
            {
                if (!pos.capture_or_promotion(ttMove))
                    update_quiet_stats(pos, ss, ttMove, stat_bonus(depth));

                // Extra penalty for early quiet moves of the previous ply
                if ((ss-1)->moveCount <= 2 && !priorCapture)
                    update_continuation_histories(ss-1, pos.piece_on(prevSq), prevSq, -stat_bonus(depth + 1));
            }
            // Penalty for a quiet ttMove that fails low
            else if (!pos.capture_or_promotion(ttMove))
            {
                int penalty = -stat_bonus(depth);
                thisThread->mainHistory[us][from_to(ttMove)] << penalty;
                update_continuation_histories(ss, pos.moved_piece(ttMove), to_sq(ttMove), penalty);
            }
        }

        if (pos.rule50_count() < 90)
            return ttValue;
    }

    // Step 5. Tablebases probe
    if (!rootNode && TB::Cardinality)
    {
#ifdef Noir
        piecesCount = popcount(pos.pieces());
#else
        piecesCount = pos.count<ALL_PIECES>();
#endif
        if (    piecesCount <= TB::Cardinality
#ifdef Add_Features //MB less probing with 7 MAN EGTB
		  &&  (piecesCount < TB::Cardinality
		  || (depth >= TB::ProbeDepth && (TB::Cardinality < 7 || TB::SevenManProbe)))
#else
		        && (piecesCount < TB::Cardinality || depth >= TB::ProbeDepth)
#endif
                &&  pos.rule50_count() == 0
                && !pos.can_castle(ANY_CASTLING))
        {
            TB::ProbeState err;

#ifdef Noir
            TB::WDLScore v = Tablebases::probe_wdl(pos, &err);

            // Force check of time on the next occasion
            if (thisThread == Threads.main())
                static_cast<MainThread*>(thisThread)->callsCnt = 0;

            if (err != TB::ProbeState::FAIL)
            {
                thisThread->tbHits.fetch_add(1, std::memory_order_relaxed);

                int drawScore = TB::UseRule50 ? 1 : 0;

                int centiPly = PawnValueEg * ss->ply / 100;

                if (    abs(v) <= drawScore
                    || !ttHit
                    || (v < -drawScore && ttValue > -VALUE_TB_WIN + centiPly + PawnValueEg * popcount(pos.pieces( pos.side_to_move())))
                    || (v >  drawScore && ttValue <  VALUE_TB_WIN - centiPly - PawnValueEg * popcount(pos.pieces(~pos.side_to_move()))))
                {
                    value =  v < -drawScore ? -VALUE_TB_WIN + centiPly + PawnValueEg * popcount(pos.pieces( pos.side_to_move()))
                           : v >  drawScore ?  VALUE_TB_WIN - centiPly - PawnValueEg * popcount(pos.pieces(~pos.side_to_move()))
                                            :  VALUE_DRAW - v < 0 ? 2 * Eval::Tempo : VALUE_ZERO;

                    tte->save(posKey, value, ttPv,
                              v > drawScore ? BOUND_LOWER : v < -drawScore ? BOUND_UPPER : BOUND_EXACT,
                              depth, MOVE_NONE, VALUE_NONE);

                    if (abs(v) <= drawScore)
                        return value;
                }
#else

            TB::WDLScore wdl = Tablebases::probe_wdl(pos, &err);

            // Force check of time on the next occasion
            if (thisThread == Threads.main())
                static_cast<MainThread*>(thisThread)->callsCnt = 0;

            if (err != TB::ProbeState::FAIL)
            {
                thisThread->tbHits.fetch_add(1, std::memory_order_relaxed);

                int drawScore = TB::UseRule50 ? 1 : 0;

                value =  wdl < -drawScore ? -VALUE_MATE + MAX_PLY + ss->ply + 1
                       : wdl >  drawScore ?  VALUE_MATE - MAX_PLY - ss->ply - 1
                                          :  VALUE_DRAW + 2 * wdl * drawScore;

                Bound b =  wdl < -drawScore ? BOUND_UPPER
                         : wdl >  drawScore ? BOUND_LOWER : BOUND_EXACT;

                if (    b == BOUND_EXACT
                    || (b == BOUND_LOWER ? value >= beta : value <= alpha))
                {
                    tte->save(posKey, value_to_tt(value, ss->ply), ttPv, b,
                              std::min(MAX_PLY - 1, depth + 6),
                              MOVE_NONE, VALUE_NONE);
                    return value;
                }

                if (PvNode)
                {
                    if (b == BOUND_LOWER)
                        bestValue = value, alpha = std::max(alpha, bestValue);
                    else
                        maxValue = value;
                }
#endif
            }
        }
    }

    // Step 6. Static evaluation of the position
    if (inCheck)

    {
        ss->staticEval = eval = VALUE_NONE;
        improving = false;
        goto moves_loop;  // Skip early pruning when in check
    }
    else if (ttHit)
    {
        // Never assume anything about values stored in TT
        ss->staticEval = eval = tte->eval();
        if (eval == VALUE_NONE)
            ss->staticEval = eval = evaluate(pos);
#ifndef Noir
        if (eval == VALUE_DRAW)
#ifdef Sullivan
            eval = value_draw(depth, thisThread);
#else
            eval = value_draw(thisThread);
#endif
#endif
        // Can ttValue be used as a better position evaluation?
        if (    ttValue != VALUE_NONE
            && (tte->bound() & (ttValue > eval ? BOUND_LOWER : BOUND_UPPER)))
            eval = ttValue;
    }
    else
    {
        if ((ss-1)->currentMove != MOVE_NULL)
        {
            int bonus = -(ss-1)->statScore / 512;

            ss->staticEval = eval = evaluate(pos) + bonus;
        }
        else
            ss->staticEval = eval = -(ss-1)->staticEval + 2 * Eval::Tempo;

        tte->save(posKey, VALUE_NONE, ttPv, BOUND_NONE, DEPTH_NONE, MOVE_NONE, eval);
    }
#ifdef Noir
  if (gameCycle)
        ss->staticEval = eval = ss->staticEval * std::max(0, (100 - pos.rule50_count())) / 100;

    improving =   ss->staticEval >= (ss-2)->staticEval
               || (ss-2)->staticEval == VALUE_NONE;

    // Begin early pruning.
    if (   !PvNode
        && !excludedMove
        && !gameCycle
        &&  abs(eval) < 2 * VALUE_KNOWN_WIN)
    {
       // Step 7. Razoring (~2 Elo)
       if (   depth < 2
           &&  ss->ply > 2 * thisThread->rootDepth / 3
           && eval <= alpha - RazorMargin)
       {
           Value q = qsearch<NonPV>(pos, ss, alpha, beta);

           if (q <= alpha)
               return q;
       }

       // Step 8. Futility pruning: child node (~30 Elo)
       if (    depth < 7
           &&  !thisThread->nmpGuard
           &&  eval - futility_margin(depth, improving) >= beta
           &&  eval < VALUE_KNOWN_WIN) // Do not return unproven wins
           return eval;

       // Step 9. Null move search with verification search (~40 Elo)
       if (   (ss-1)->currentMove != MOVE_NULL
           && (ss-1)->statScore < 22661
           &&  eval >= beta
           &&  eval >= ss->staticEval
           &&  ss->staticEval >= beta - 33 * depth + 299 - improving * 30
           &&  pos.non_pawn_material(us)
           && !thisThread->nmpGuard
           && !(depth > 4 && (MoveList<LEGAL, KING>(pos).size() < 1 || MoveList<LEGAL>(pos).size() < 6)))
#else
#ifndef Noir
#ifdef Fortress
      if (gameCycle)
          ss->staticEval = eval = ss->staticEval * std::max(0, (100 - pos.rule50_count())) / 100;
#endif
    // Step 7. Razoring (~0 Elo)
    if (   !rootNode // The required rootNode PV handling is not available in qsearch
#ifdef Weakfish
        && !weakFishSearch
#endif
        &&  depth < 2
#ifdef Fortress
        &&  !gameCycle
#endif
        &&  eval <= alpha - RazorMargin)
        return qsearch<NT>(pos, ss, alpha, beta);

    improving =  (ss-2)->staticEval == VALUE_NONE ? (ss->staticEval >= (ss-4)->staticEval
              || (ss-4)->staticEval == VALUE_NONE) : ss->staticEval >= (ss-2)->staticEval;

    // Step 8. Futility pruning: child node (~50 Elo)
    if (   !PvNode
#ifdef Weakfish
        && !weakFishSearch
#endif
		&&  depth < 6
#ifdef Fortress
        &&  !gameCycle
#endif
        &&  eval - futility_margin(depth, improving) >= beta
        &&  eval < VALUE_KNOWN_WIN) // Do not return unproven wins
        return eval;

    // Step 9. Null move search with verification search (~39 Elo)
    if (   !PvNode
        && (ss-1)->currentMove != MOVE_NULL

	      && (ss-1)->statScore < 23397
	      &&  eval >= beta
        &&  eval >= ss->staticEval
        &&  ss->staticEval >= beta - 32 * depth + 292 - improving * 30
        && !excludedMove
#if defined (Sullivan) || (Bluefish)  //authored by Jörg Oster originally, in corchess by Ivan Ilvec
        && thisThread->selDepth + 3 > thisThread->rootDepth
#endif
        &&  pos.non_pawn_material(us)
        && (ss->ply >= thisThread->nmpMinPly || us != thisThread->nmpColor))
#endif
#endif
    {
        assert(eval - beta >= 0);

        // Null move dynamic reduction based on depth and value
        Depth R = (854 + 68 * depth) / 258 + std::min(int(eval - beta) / 192, 3);

        ss->currentMove = MOVE_NULL;
        ss->continuationHistory = &thisThread->continuationHistory[0][0][NO_PIECE][0];

        pos.do_null_move(st);
        Value nullValue = -search<NonPV>(pos, ss+1, -beta, -beta+1, depth-R, !cutNode);
        pos.undo_null_move();

        if (nullValue >= beta)
        {
            // Do not return unproven mate scores
            if (nullValue >= VALUE_MATE_IN_MAX_PLY)
                nullValue = beta;
#ifdef Noir
              if (abs(beta) < VALUE_KNOWN_WIN && depth < 11)
                   return nullValue;

               // Do verification search at high depths
               thisThread->nmpGuard = true;

               Value v = search<NonPV>(pos, ss, beta-1, beta, depth-R, false);

               thisThread->nmpGuard = false;
#else

            if (thisThread->nmpMinPly || (abs(beta) < VALUE_KNOWN_WIN && depth < 13))
                return nullValue;

            assert(!thisThread->nmpMinPly); // Recursive verification is not allowed

            // Do verification search at high depths, with null move pruning disabled
            // for us, until ply exceeds nmpMinPly.
            thisThread->nmpMinPly = ss->ply + 3 * (depth-R) / 4;
            thisThread->nmpColor = us;

            Value v = search<NonPV>(pos, ss, beta-1, beta, depth-R, false);

            thisThread->nmpMinPly = 0;
#endif
            if (v >= beta)
                return nullValue;
        }
    }

    // Step 10. ProbCut (~10 Elo)
    // If we have a good enough capture and a reduced search returns a value
    // much above beta, we can (almost) safely prune the previous move.
#ifdef Noir
       if (    depth >= 5
           &&  ss->ply % 2 == 0
           &&  !thisThread->nmpGuard
           &&  abs(beta) < VALUE_MATE_IN_MAX_PLY)
       {
           Value raisedBeta = std::min(beta + 191 - 46 * improving, VALUE_INFINITE);
           MovePicker mp(pos, ttMove, raisedBeta - ss->staticEval, &thisThread->captureHistory);
           int probCutCount = 0;

           while (  (move = mp.next_move()) != MOVE_NONE
                  && probCutCount < 2 + 2 * cutNode)
               if (move != excludedMove)
#else
    if (   !PvNode
#ifdef Weakfish
        && !weakFishSearch
#endif
        &&  depth >= 5
        &&  abs(beta) < VALUE_MATE_IN_MAX_PLY)
    {
        Value raisedBeta = std::min(beta + 189 - 45 * improving, VALUE_INFINITE);
        MovePicker mp(pos, ttMove, raisedBeta - ss->staticEval, &thisThread->captureHistory);
        int probCutCount = 0;

        while (  (move = mp.next_move()) != MOVE_NONE
               && probCutCount < 2 + 2 * cutNode)
            if (move != excludedMove && pos.legal(move))
#endif
            {
                assert(pos.capture_or_promotion(move));
                assert(depth >= 5);

                captureOrPromotion = true;
                probCutCount++;

                ss->currentMove = move;
                ss->continuationHistory = &thisThread->continuationHistory[inCheck]
                                                                          [captureOrPromotion]
                                                                          [pos.moved_piece(move)]
                                                                          [to_sq(move)];

                pos.do_move(move, st);

                // Perform a preliminary qsearch to verify that the move holds
                value = -qsearch<NonPV>(pos, ss+1, -raisedBeta, -raisedBeta+1);

                // If the qsearch held, perform the regular search
                if (value >= raisedBeta)
                    value = -search<NonPV>(pos, ss+1, -raisedBeta, -raisedBeta+1, depth - 4, !cutNode);

                pos.undo_move(move);

                   if (value >= raisedBeta)
                       return value;
               }
       }
#ifdef Noir
    } //End early Pruning
#endif

    // Step 11. Internal iterative deepening (~1 Elo)
    if (depth >= 7 && !ttMove)
    {
        search<NT>(pos, ss, alpha, beta, depth - 7, cutNode);

#if defined (Sullivan) || (Blau) || (Fortress)
        tte = probeTT(pos, ss, posKey, ttHit, ttValue, ttMove);
#else
        tte = TT.probe(posKey, ttHit);
#endif
        ttValue = ttHit ? value_from_tt(tte->value(), ss->ply, pos.rule50_count()) : VALUE_NONE;
        ttMove = ttHit ? tte->move() : MOVE_NONE;
    }

moves_loop: // When in check, search starts from here

    const PieceToHistory* contHist[] = { (ss-1)->continuationHistory, (ss-2)->continuationHistory,
                                          nullptr                   , (ss-4)->continuationHistory,
                                          nullptr                   , (ss-6)->continuationHistory };

    Move countermove = thisThread->counterMoves[pos.piece_on(prevSq)][prevSq];

    MovePicker mp(pos, ttMove, depth, &thisThread->mainHistory,
                                      &thisThread->captureHistory,
                                      contHist,
                                      countermove,
                                      ss->killers);

    value = bestValue;
    singularLMR = moveCountPruning = false;
    ttCapture = ttMove && pos.capture_or_promotion(ttMove);

    // Mark this node as being searched
    ThreadHolding th(thisThread, posKey, ss->ply);

    // Step 12. Loop through all pseudo-legal moves until no moves remain
    // or a beta cutoff occurs.
    while ((move = mp.next_move(moveCountPruning)) != MOVE_NONE)
    {
      assert(is_ok(move));

      if (move == excludedMove)
          continue;

      // At root obey the "searchmoves" option and skip moves not listed in Root
      // Move List. As a consequence any illegal move is also skipped. In MultiPV
      // mode we also skip PV moves which have been already searched and those
      // of lower "TB rank" if we are in a TB root position.
      if (rootNode && !std::count(thisThread->rootMoves.begin() + thisThread->pvIdx,
                                  thisThread->rootMoves.begin() + thisThread->pvLast, move))
          continue;

      ss->moveCount = ++moveCount;

#ifdef Add_Features
      if (!minOutput && rootNode && thisThread == Threads.main() && Time.elapsed() > 3000)
#else
      if (rootNode && thisThread == Threads.main() && Time.elapsed() > 3000)
#endif
          sync_cout << "info depth " << depth
                    << " currmove " << UCI::move(move, pos.is_chess960())
                    << " currmovenumber " << moveCount + thisThread->pvIdx << sync_endl;
      if (PvNode)
          (ss+1)->pv = nullptr;

      extension = 0;
      captureOrPromotion = pos.capture_or_promotion(move);
      movedPiece = pos.moved_piece(move);
      givesCheck = pos.gives_check(move);
#ifdef Noir
      isMate = false;

      if (givesCheck)
      {
          pos.do_move(move, st, givesCheck);
          isMate = MoveList<LEGAL>(pos).size() == 0;
          pos.undo_move(move);

          if (!isMate) // Don't double count nodes
              thisThread->nodes.fetch_sub(1, std::memory_order_relaxed);
      }

      if (isMate)
      {
          ss->currentMove = move;
          ss->continuationHistory = &thisThread->continuationHistory[inCheck][priorCapture][movedPiece][to_sq(move)];
          value = mate_in(ss->ply+1);

          if (PvNode && (moveCount == 1 || (value > alpha && (rootNode || value < beta))))
          {
              (ss+1)->pv = pv;
              (ss+1)->pv[0] = MOVE_NONE;
          }
      }
      else
      {
#endif

		// Calculate new depth for this move
                newDepth = depth - 1;
#if defined (Fortress) || (Noir)
                // Step 13. Pruning at shallow depth (~204 Elo)
                if (  !PvNode
#else
		// Step 13. Pruning at shallow depth (~204 Elo)
		if (  !rootNode
#endif
#ifdef Weakfish
			&& !weakFishSearch
#endif
			&& pos.non_pawn_material(us)
			&& bestValue > VALUE_MATED_IN_MAX_PLY)
		{
			// Skip quiet moves if movecount exceeds our FutilityMoveCount threshold
			moveCountPruning = moveCount >= futility_move_count(improving, depth);

			if (
				!captureOrPromotion
				&& !givesCheck
#if defined (Sullivan) || (Blau) || (Fortress) || (Noir)
                && (!PvNode || !pos.advanced_pawn_push(move) || pos.non_pawn_material(~us) > BishopValueMg)
#endif
                )
			{
				// Reduced depth of the next LMR search
				int lmrDepth = std::max(newDepth - reduction(improving, depth, moveCount), 0);

				// Countermoves based pruning (~20 Elo)
				if (   lmrDepth < 4 + ((ss-1)->statScore > 0 || (ss-1)->moveCount == 1)
					&& (*contHist[0])[movedPiece][to_sq(move)] < CounterMovePruneThreshold
					&& (*contHist[1])[movedPiece][to_sq(move)] < CounterMovePruneThreshold)
					continue;

				// Futility pruning: parent node (~2 Elo)
#if defined (Fortress) || (Noir)
                if (   lmrDepth < 3
#else
                if (   lmrDepth < 6
#endif
                    && !inCheck
                    && ss->staticEval + 235 + 172 * lmrDepth <= alpha
                    &&  thisThread->mainHistory[us][from_to(move)]
                    + (*contHist[0])[movedPiece][to_sq(move)]
                    + (*contHist[1])[movedPiece][to_sq(move)]
                    + (*contHist[3])[movedPiece][to_sq(move)] < 25000)
                         continue;
                    // Prune moves with negative SEE (~20 Elo)
                    if (!pos.see_ge(move, Value(-(32 - std::min(lmrDepth, 18)) * lmrDepth * lmrDepth)))
                             continue;
             }
                    else if (!pos.see_ge(move, Value(-194) * depth)) // (~20 Elo)
                        continue;

		}

      // Step 14. Extensions (~74 Elo)

#if defined (Fortress) || (Noir)
      if (   gameCycle
          && (depth < 5 || PvNode))
              extension = (2 - (ss->ply % 2 == 0 && !PvNode));
#endif
      // Singular extension search (~67 Elo). If all moves but one fail low on a
      // search of (alpha-s, beta-s), and just one fails high on (alpha, beta),
      // then that move is singular and should be extended. To verify this we do
      // a reduced search on all the other moves but the ttMove and if the
      // result is lower than ttValue minus a margin then we will extend the ttMove.
#if defined (Fortress) || (Noir)
      else if (    depth >= 6
#else
      if (    depth >= 6
#endif
          &&  move == ttMove
#if defined (Fortress) || (Noir)
          &&  !gameCycle
#endif
          && !rootNode

          && !excludedMove // Avoid recursive singular search
#if defined (Fortress) || (Noir)
          &&  ttValue != VALUE_NONE //Already implicit in the next condition */
#else
          &&  abs(ttValue) < VALUE_KNOWN_WIN
#endif
          && (tte->bound() & BOUND_LOWER)
          &&  tte->depth() >= depth - 3
#ifndef Noir
          &&  pos.legal(move)
#endif
		  )
      {
#if defined (Fortress) || (Noir)
          Value singularBeta = std::max(ttValue - 2 * depth, mated_in(ss->ply));
#else
          Value singularBeta = ttValue - 2 * depth;
#endif
          Depth halfDepth = depth / 2;
          ss->excludedMove = move;
          value = search<NonPV>(pos, ss, singularBeta - 1, singularBeta, halfDepth, cutNode);
          ss->excludedMove = MOVE_NONE;
          if (value < singularBeta)
          {
              extension = 1;
              singularLMR = true;
          }

          // Multi-cut pruning
          // Our ttMove is assumed to fail high, and now we failed high also on a reduced
          // search without the ttMove. So we assume this expected Cut-node is not singular,

          // that multiple moves fail high, and we can prune the whole subtree by returning
          // a soft bound.
#if defined (Stockfish) || (Weakfish)
          else if (singularBeta >= beta)
#else
          else if (   eval >= beta
                   && singularBeta >= beta)
#endif
              return singularBeta;
      }

      // Check extension (~2 Elo)

#if defined (Sullivan) || (Blau) || (Fortress) || (Noir)

       else if (    givesCheck
               && (pos.is_discovery_check_on_king(~us, move) || pos.see_ge(move))
#ifndef Noir
               && ++thisThread->extension < thisThread->nodes.load(std::memory_order_relaxed) /4 //MichaelB7
#endif
                )
               extension = 1;

     // MichaelB7 Passed pawn extension
      else if ( move == ss->killers[0] && !extension
			  && (pos.promotion_pawn_push(move)
                          || (pos.advanced_pawn_push(move)
                          && pos.pawn_passed(us, to_sq(move)))))
          extension = 1;

#else
      else if (    givesCheck
               && (pos.is_discovery_check_on_king(~us, move) || pos.see_ge(move)))
          extension = 1;

      // Passed pawn extension
      else if (   move == ss->killers[0]
               && pos.advanced_pawn_push(move)
               && pos.pawn_passed(us, to_sq(move)))
          extension = 1;
#endif

      // Last captures extension
      else if (   PieceValue[EG][pos.captured_piece()] > PawnValueEg
               && pos.non_pawn_material() <= 2 * RookValueMg)
          extension = 1;

      // Castling extension
      if (type_of(move) == CASTLING)
          extension = 1;

      // Add extension to new depth
      newDepth += extension;

      // Speculative prefetch as early as possible
      prefetch(TT.first_entry(pos.key_after(move)));

      // Check for legality just before making the move
#ifndef Noir
      if (!rootNode && !pos.legal(move))
      {
          ss->moveCount = --moveCount;
          continue;
      }
#endif
      // Update the current move (this must be done after singular extension search)
      ss->currentMove = move;
      ss->continuationHistory = &thisThread->continuationHistory[inCheck]
                                                                [captureOrPromotion]
                                                                [movedPiece]
                                                                [to_sq(move)];

      // Step 15. Make the move
      pos.do_move(move, st, givesCheck);

      // Step 16. Reduced depth search (LMR, ~210 Elo). If the move fails high it will be
      // re-searched at full depth..
#ifdef Weakfish
      if (    !weakFishSearch && depth >= 3
#else
      if (    depth >= 3
#endif
#if defined (Fortress) || (Noir)
          &&  !gameCycle
#endif
          &&  moveCount > 1 + rootNode + (rootNode && bestValue < alpha)
          && (!rootNode || thisThread->best_move_count(move) == 0)
#if defined (Fortress) || (Noir)
          &&  thisThread->selDepth > depth
#endif
          && (  !captureOrPromotion
              || moveCountPruning
              || ss->staticEval + PieceValue[EG][pos.captured_piece()] <= alpha
              || cutNode
#ifndef Noir
#if defined (Stockfish) || (Weakfish)
              || thisThread->ttHitAverage < 375 * ttHitAverageResolution * ttHitAverageWindow / 1024
#else
			  || thisThread->ttHitAverage < 384 * ttHitAverageResolution * ttHitAverageWindow / 1024
#endif
#endif
      ))
      {
           Depth r = reduction(improving, depth, moveCount);
#ifndef Noir
#if defined (Stockfish) || (Weakfish)
           if (thisThread->ttHitAverage > 500 * ttHitAverageResolution * ttHitAverageWindow / 1024)
#else
           // Decrease reduction if the ttHit running average is large
           if (thisThread->ttHitAverage > 544 * ttHitAverageResolution * ttHitAverageWindow / 1024)
#endif
               r--;
#endif

          // Reduction if other threads are searching this position.
          if (th.marked())
              r++;

          // Decrease reduction if position is or has been on the PV (~8 Elo)
          if (ttPv)
              r -= 2;

          // Decrease reduction if opponent's move count is high (~5 Elo)
          if ((ss-1)->moveCount > 14)
              r--;

          // Decrease reduction if ttMove has been singularly extended (~3 Elo)
          if (singularLMR)
              r -= 2;
#if defined (Fortress) || (Noir)
          if (!PvNode && !captureOrPromotion)
#else
          if (!captureOrPromotion)
#endif
          {
              // Increase reduction if ttMove is a capture (~5 Elo)
              if (ttCapture)
                  r++;

              // Increase reduction for cut nodes (~10 Elo)
              if (cutNode)
                  r += 2;

              // Decrease reduction for moves that escape a capture. Filter out
              // castling moves, because they are coded as "king captures rook" and
              // hence break make_move(). (~2 Elo)
              else if (    type_of(move) == NORMAL
                       && !pos.see_ge(reverse_move(move)))
                  r -= 2;

              ss->statScore =  thisThread->mainHistory[us][from_to(move)]
                             + (*contHist[0])[movedPiece][to_sq(move)]
                             + (*contHist[1])[movedPiece][to_sq(move)]
                             + (*contHist[3])[movedPiece][to_sq(move)]
                             - 4926;

              // Reset statScore to zero if negative and most stats shows >= 0
              if (    ss->statScore < 0
                  && (*contHist[0])[movedPiece][to_sq(move)] >= 0
                  && (*contHist[1])[movedPiece][to_sq(move)] >= 0
                  && thisThread->mainHistory[us][from_to(move)] >= 0)
                  ss->statScore = 0;

              // Decrease/increase reduction by comparing opponent's stat score (~10 Elo)
              if (ss->statScore >= -102 && (ss-1)->statScore < -114)
                  r--;
              else if ((ss-1)->statScore >= -116 && ss->statScore < -154)
                  r++;

              // Decrease/increase reduction for moves with a good/bad history (~26 Elo)
              r -= ss->statScore / 16384;
          }

          // Increase reduction for captures/promotions if late move and at low depth
          else if (depth < 8 && moveCount > 2)
              r++;

#if defined (Fortress) || (Noir)
          Depth rr = newDepth / (2 + ss->ply / 3);
          r -= rr;
#endif

          Depth d = clamp(newDepth - r, 1, newDepth);

          value = -search<NonPV>(pos, ss+1, -(alpha+1), -alpha, d, true);

          doFullDepthSearch = (value > alpha && d != newDepth), didLMR = true;
      }
      else
          doFullDepthSearch = !PvNode || moveCount > 1, didLMR = false;

      // Step 17. Full depth search when LMR is skipped or fails high
      if (doFullDepthSearch)
      {
          value = -search<NonPV>(pos, ss+1, -(alpha+1), -alpha, newDepth, !cutNode);

          if (didLMR && !captureOrPromotion)
          {
              int bonus = value > alpha ?  stat_bonus(newDepth)
                                        : -stat_bonus(newDepth);

              if (move == ss->killers[0])
                  bonus += bonus / 4;

              update_continuation_histories(ss, movedPiece, to_sq(move), bonus);
          }
      }

      // For PV nodes only, do a full PV search on the first move or after a fail
      // high (in the latter case search only if value < beta), otherwise let the
      // parent node fail low with value <= alpha and try another move.
      if (PvNode && (moveCount == 1 || (value > alpha && (rootNode || value < beta))))
      {
          (ss+1)->pv = pv;
          (ss+1)->pv[0] = MOVE_NONE;

          value = -search<PV>(pos, ss+1, -beta, -alpha, newDepth, false);
      }

      // Step 18. Undo move
      pos.undo_move(move);
#ifdef Noir
   }
#endif

      assert(value > -VALUE_INFINITE && value < VALUE_INFINITE);

      // Step 19. Check for a new best move
      // Finished searching the move. If a stop occurred, the return value of
      // the search cannot be trusted, and we return immediately without
      // updating best move, PV and TT.
      if (Threads.stop.load(std::memory_order_relaxed))
          return VALUE_ZERO;

      if (rootNode)
      {
          RootMove& rm = *std::find(thisThread->rootMoves.begin(),
                                    thisThread->rootMoves.end(), move);

          // PV move or new best move?
          if (moveCount == 1 || value > alpha)
          {
              rm.score = value;
              rm.selDepth = thisThread->selDepth;
              rm.pv.resize(1);

              assert((ss+1)->pv);

              for (Move* m = (ss+1)->pv; *m != MOVE_NONE; ++m)
                  rm.pv.push_back(*m);

              // We record how often the best move has been changed in each
              // iteration. This information is used for time management: When
              // the best move changes frequently, we allocate some more time.

			  if (moveCount > 1)
				  ++thisThread->bestMoveChanges;

          }
          else
              // All other moves but the PV are set to the lowest value: this
              // is not a problem when sorting because the sort is stable and the
              // move position in the list is preserved - just the PV is pushed up.
              rm.score = -VALUE_INFINITE;
      }

      if (value > bestValue)
      {
          bestValue = value;

          if (value > alpha)
          {
              bestMove = move;

              if (PvNode && !rootNode) // Update pv even in fail-high case
                  update_pv(ss->pv, move, (ss+1)->pv);

              if (PvNode && value < beta) // Update alpha! Always alpha < beta
                  alpha = value;
              else
              {
                  assert(value >= beta); // Fail high
                  ss->statScore = 0;
				   break;
              }
          }
      }

      if (move != bestMove)
      {
          if (captureOrPromotion && captureCount < 32)
              capturesSearched[captureCount++] = move;

          else if (!captureOrPromotion && quietCount < 64)
              quietsSearched[quietCount++] = move;
      }
    }

    // The following condition would detect a stop only after move loop has been
    // completed. But in this case bestValue is valid because we have fully
    // searched our subtree, and we can anyhow save the result in TT.
    /*
       if (Threads.stop)
        return VALUE_DRAW;
    */

    // Step 20. Check for mate and stalemate
    // All legal moves have been searched and if there are no legal moves, it
    // must be a mate or a stalemate. If we are in a singular extension search then
    // return a fail low score.

    assert(moveCount || !inCheck || excludedMove || !MoveList<LEGAL>(pos).size());

    if (!moveCount)
        bestValue = excludedMove ? alpha
                   :     inCheck ? mated_in(ss->ply) : VALUE_DRAW;

    else if (bestMove)
        update_all_stats(pos, ss, bestMove, bestValue, beta, prevSq,
                         quietsSearched, quietCount, capturesSearched, captureCount, depth);

    // Bonus for prior countermove that caused the fail low
    else if (   (depth >= 3 || PvNode)
             && !priorCapture)
        update_continuation_histories(ss-1, pos.piece_on(prevSq), prevSq, stat_bonus(depth));
#ifndef Noir
    if (PvNode)
        bestValue = std::min(bestValue, maxValue);
#endif
    if (!excludedMove)
        tte->save(posKey, value_to_tt(bestValue, ss->ply), ttPv,
                  bestValue >= beta ? BOUND_LOWER :
                  PvNode && bestMove ? BOUND_EXACT : BOUND_UPPER,
#if defined (Sullivan) || (Blau) || (Fortress) || (Noir)
                  std::min(depth + 6 * int(pos.rule50_count() > 90), MAX_PLY - 1),
                       bestMove, ss->staticEval); // joergoster patch
#else
                  depth, bestMove, ss->staticEval);
#endif

    assert(bestValue > -VALUE_INFINITE && bestValue < VALUE_INFINITE);

    return bestValue;
  }

  // qsearch() is the quiescence search function, which is called by the main search
  // function with zero depth, or recursively with further decreasing depth per call.
  template <NodeType NT>
  Value qsearch(Position& pos, Stack* ss, Value alpha, Value beta, Depth depth) {

    constexpr bool PvNode = NT == PV;

    assert(alpha >= -VALUE_INFINITE && alpha < beta && beta <= VALUE_INFINITE);
    assert(PvNode || (alpha == beta - 1));
    assert(depth <= 0);

    Move pv[MAX_PLY+1];
    StateInfo st;
    TTEntry* tte;
    Key posKey;
    Move ttMove, move, bestMove;
    Depth ttDepth;
    Value bestValue, value, ttValue, futilityValue, futilityBase, oldAlpha;

#if defined (Fortress) || (Noir)
    bool ttHit, pvHit, inCheck, givesCheck, captureOrPromotion, evasionPrunable, gameCycle;
#else
    bool ttHit, pvHit, inCheck, givesCheck, captureOrPromotion, evasionPrunable;
#endif

    int moveCount;

    if (PvNode)
    {
        oldAlpha = alpha; // To flag BOUND_EXACT when eval above alpha and no available moves
        (ss+1)->pv = pv;
        ss->pv[0] = MOVE_NONE;
    }

    Thread* thisThread = pos.this_thread();
    (ss+1)->ply = ss->ply + 1;
    bestMove = MOVE_NONE;
    inCheck = pos.checkers();
    moveCount = 0;
#if defined (Fortress) || (Noir)
    gameCycle = false;

    if (pos.has_game_cycle(ss->ply))
    {
       if (VALUE_DRAW >= beta)
           return VALUE_DRAW;

       alpha = std::max(alpha, VALUE_DRAW);
       gameCycle = true;
    }

    if (pos.is_draw(ss->ply))
        return VALUE_DRAW;

    // Check for an immediate draw or maximum ply reached
    if (ss->ply >= MAX_PLY)
        return !inCheck ? evaluate(pos) : VALUE_DRAW;

    if (alpha >= mate_in(ss->ply+1))
        return alpha;
#else
    // Check for an immediate draw or maximum ply reached
    if (   pos.is_draw(ss->ply)
        || ss->ply >= MAX_PLY)
        return (ss->ply >= MAX_PLY && !inCheck) ? evaluate(pos) : VALUE_DRAW;
#endif
    assert(0 <= ss->ply && ss->ply < MAX_PLY);

    // Decide whether or not to include checks: this fixes also the type of
    // TT entry depth that we are going to use. Note that in qsearch we use
    // only two types of depth in TT: DEPTH_QS_CHECKS or DEPTH_QS_NO_CHECKS.
    ttDepth = inCheck || depth >= DEPTH_QS_CHECKS ? DEPTH_QS_CHECKS
                                                  : DEPTH_QS_NO_CHECKS;
    // Transposition table lookup
    posKey = pos.key();
#if defined (Sullivan) || (Blau) || (Fortress)
   tte = probeTT(pos, ss, posKey, ttHit, ttValue, ttMove);
   int piecesCountqs = pos.count<ALL_PIECES>();
#else
    tte = TT.probe(posKey, ttHit);
#endif
    ttValue = ttHit ? value_from_tt(tte->value(), ss->ply, pos.rule50_count()) : VALUE_NONE;
    ttMove = ttHit ? tte->move() : MOVE_NONE;
    pvHit = ttHit && tte->is_pv();

    if (  !PvNode
        && ttHit
#if defined (Sullivan) || (Blau) || (Fortress)
        && (pos.rule50_count() < 92 || (piecesCountqs < 8  && TB::Cardinality ))
#endif
#if defined (Fortress) || (Noir)
        &&  !gameCycle
#endif
        && tte->depth() >= ttDepth
        && ttValue != VALUE_NONE // Only in case of TT access race
#if defined (Fortress) || (Noir)
        && (ttValue != VALUE_DRAW || VALUE_DRAW >= beta)
#endif
        && (ttValue >= beta ? (tte->bound() & BOUND_LOWER)
                            : (tte->bound() & BOUND_UPPER)))
        return ttValue;

    // Evaluate the position statically
    if (inCheck)
    {
        ss->staticEval = VALUE_NONE;
        bestValue = futilityBase = -VALUE_INFINITE;
    }
    else
    {
        if (ttHit)
        {
            // Never assume anything about values stored in TT
            if ((ss->staticEval = bestValue = tte->eval()) == VALUE_NONE)
                ss->staticEval = bestValue = evaluate(pos);

            // Can ttValue be used as a better position evaluation?
            if (    ttValue != VALUE_NONE
                && (tte->bound() & (ttValue > bestValue ? BOUND_LOWER : BOUND_UPPER)))
                bestValue = ttValue;
        }
        else
            ss->staticEval = bestValue =
            (ss-1)->currentMove != MOVE_NULL ? evaluate(pos)
                                             : -(ss-1)->staticEval + 2 * Eval::Tempo;

        // Stand pat. Return immediately if static value is at least beta
        if (bestValue >= beta)
        {
            if (!ttHit)
                tte->save(posKey, value_to_tt(bestValue, ss->ply), pvHit, BOUND_LOWER,
                          DEPTH_NONE, MOVE_NONE, ss->staticEval);

            return bestValue;
        }

        if (PvNode && bestValue > alpha)
            alpha = bestValue;

        futilityBase = bestValue + 154;
    }
#if defined (Fortress) || (Noir)
    if (gameCycle && !inCheck)
        ss->staticEval = bestValue = ss->staticEval * std::max(0, (100 - pos.rule50_count())) / 100;
#endif
    const PieceToHistory* contHist[] = { (ss-1)->continuationHistory, (ss-2)->continuationHistory,
                                          nullptr                   , (ss-4)->continuationHistory,
                                          nullptr                   , (ss-6)->continuationHistory };

    // Initialize a MovePicker object for the current position, and prepare
    // to search the moves. Because the depth is <= 0 here, only captures,
    // queen promotions and checks (only if depth >= DEPTH_QS_CHECKS) will
    // be generated.
    MovePicker mp(pos, ttMove, depth, &thisThread->mainHistory,
                                      &thisThread->captureHistory,
                                      contHist,
                                      to_sq((ss-1)->currentMove));

    // Loop through the moves until no moves remain or a beta cutoff occurs
    while ((move = mp.next_move()) != MOVE_NONE)
    {
      assert(is_ok(move));

      givesCheck = pos.gives_check(move);
      captureOrPromotion = pos.capture_or_promotion(move);

      moveCount++;

#ifdef Noir
      if (!PvNode)
      {
#endif
      // Futility pruning
      if (   !inCheck
          && !givesCheck
          &&  futilityBase > -VALUE_KNOWN_WIN
          && !pos.advanced_pawn_push(move))
      {
          assert(type_of(move) != ENPASSANT); // Due to !pos.advanced_pawn_push

          futilityValue = futilityBase + PieceValue[EG][pos.piece_on(to_sq(move))];

          if (futilityValue <= alpha)
          {
              bestValue = std::max(bestValue, futilityValue);
              continue;
          }

          if (futilityBase <= alpha && !pos.see_ge(move, VALUE_ZERO + 1))
          {
              bestValue = std::max(bestValue, futilityBase);
              continue;
          }
      }

      // Detect non-capture evasions that are candidates to be pruned
      evasionPrunable =    inCheck
                       &&  (depth != 0 || moveCount > 2)
                       &&  bestValue > VALUE_MATED_IN_MAX_PLY
                       && !pos.capture(move);

      // Don't search moves with negative SEE values
#if defined (Stockfish) || (Weakfish)
      if (  (!inCheck || evasionPrunable) && !pos.see_ge(move))
#else
      if (  (!inCheck || evasionPrunable)
          && !(givesCheck && pos.is_discovery_check_on_king(~pos.side_to_move(), move))
          && !pos.see_ge(move))
#endif
          continue;
#ifdef Noir
      }
#endif

      // Speculative prefetch as early as possible
      prefetch(TT.first_entry(pos.key_after(move)));
#ifndef Noir
      // Check for legality just before making the move
      if (!pos.legal(move))
      {
          moveCount--;
          continue;
      }
#endif
      ss->currentMove = move;
      ss->continuationHistory = &thisThread->continuationHistory[inCheck]
                                                                [captureOrPromotion]
                                                                [pos.moved_piece(move)]
                                                                [to_sq(move)];

      // Make and search the move
      pos.do_move(move, st, givesCheck);
      value = -qsearch<NT>(pos, ss+1, -beta, -alpha, depth - 1);
      pos.undo_move(move);

      assert(value > -VALUE_INFINITE && value < VALUE_INFINITE);

      // Check for a new best move
      if (value > bestValue)
      {
          bestValue = value;

          if (value > alpha)
          {
              bestMove = move;

              if (PvNode) // Update pv even in fail-high case
                  update_pv(ss->pv, move, (ss+1)->pv);

              if (PvNode && value < beta) // Update alpha here!
                  alpha = value;
              else
                  break; // Fail high
          }
       }
    }

#if defined (Add_Features)
	  if (!adaptive && jekyll && (bestValue + (255 * PawnValueEg / (uci_elo/10)) >= 0 ))
        {
			//int o_value = bestValue;// for debug
			//sync_cout << "Value " << bestValue << sync_endl;// for debug
			bestValue += (rand() % 64 * 2001/uci_elo + 1);
			//sync_cout << "Random Value " << bestValue << sync_endl;// for debug
			//sync_cout << "Change " << bestValue - o_value << sync_endl;// for debug
	  }
#endif

    // All legal moves have been searched. A special case: If we're in check
    // and no legal moves were found, it is checkmate.
    if (inCheck && bestValue == -VALUE_INFINITE)
        return mated_in(ss->ply); // Plies to mate from the root

    tte->save(posKey, value_to_tt(bestValue, ss->ply), pvHit,
              bestValue >= beta ? BOUND_LOWER :
              PvNode && bestValue > oldAlpha  ? BOUND_EXACT : BOUND_UPPER,
              ttDepth, bestMove, ss->staticEval);

    assert(bestValue > -VALUE_INFINITE && bestValue < VALUE_INFINITE);

    return bestValue;
  }


  // value_to_tt() adjusts a mate score from "plies to mate from the root" to
  // "plies to mate from the current position". Non-mate scores are unchanged.
  // The function is called before storing a value in the transposition table.

  Value value_to_tt(Value v, int ply) {

    assert(v != VALUE_NONE);

    return  v >= VALUE_MATE_IN_MAX_PLY  ? v + ply
          : v <= VALUE_MATED_IN_MAX_PLY ? v - ply : v;
  }


  // value_from_tt() is the inverse of value_to_tt(): It adjusts a mate score
  // from the transposition table (which refers to the plies to mate/be mated
  // from current position) to "plies to mate/be mated from the root".
  Value value_from_tt(Value v, int ply, int r50c) {
    return  v == VALUE_NONE             ? VALUE_NONE
          : v >= VALUE_MATE_IN_MAX_PLY  ? VALUE_MATE - v > 99 - r50c ? VALUE_MATE_IN_MAX_PLY  : v - ply
          : v <= VALUE_MATED_IN_MAX_PLY ? VALUE_MATE + v > 99 - r50c ? VALUE_MATED_IN_MAX_PLY : v + ply : v;
  }


  // update_pv() adds current move and appends child pv[]

  void update_pv(Move* pv, Move move, Move* childPv) {

    for (*pv++ = move; childPv && *childPv != MOVE_NONE; )
        *pv++ = *childPv++;
    *pv = MOVE_NONE;
  }


  // update_all_stats() updates stats at the end of search() when a bestMove is found

  void update_all_stats(const Position& pos, Stack* ss, Move bestMove, Value bestValue, Value beta, Square prevSq,
                        Move* quietsSearched, int quietCount, Move* capturesSearched, int captureCount, Depth depth) {

    int bonus1, bonus2;
    Color us = pos.side_to_move();
    Thread* thisThread = pos.this_thread();
    CapturePieceToHistory& captureHistory = thisThread->captureHistory;
    Piece moved_piece = pos.moved_piece(bestMove);
    PieceType captured = type_of(pos.piece_on(to_sq(bestMove)));

    bonus1 = stat_bonus(depth + 1);
    bonus2 = bestValue > beta + PawnValueMg ? bonus1               // larger bonus
                                            : stat_bonus(depth);   // smaller bonus

    if (!pos.capture_or_promotion(bestMove))
    {
        update_quiet_stats(pos, ss, bestMove, bonus2);

        // Decrease all the non-best quiet moves
        for (int i = 0; i < quietCount; ++i)
        {
            thisThread->mainHistory[us][from_to(quietsSearched[i])] << -bonus2;
            update_continuation_histories(ss, pos.moved_piece(quietsSearched[i]), to_sq(quietsSearched[i]), -bonus2);
        }
    }
    else
        captureHistory[moved_piece][to_sq(bestMove)][captured] << bonus1;

    // Extra penalty for a quiet TT or main killer move in previous ply when it gets refuted
    if (   ((ss-1)->moveCount == 1 || ((ss-1)->currentMove == (ss-1)->killers[0]))
        && !pos.captured_piece())
            update_continuation_histories(ss-1, pos.piece_on(prevSq), prevSq, -bonus1);

    // Decrease all the non-best capture moves
    for (int i = 0; i < captureCount; ++i)
    {
        moved_piece = pos.moved_piece(capturesSearched[i]);
        captured = type_of(pos.piece_on(to_sq(capturesSearched[i])));
        captureHistory[moved_piece][to_sq(capturesSearched[i])][captured] << -bonus1;
    }
  }


  // update_continuation_histories() updates histories of the move pairs formed
  // by moves at ply -1, -2, and -4 with current move.

  void update_continuation_histories(Stack* ss, Piece pc, Square to, int bonus) {

    for (int i : {1, 2, 4, 6})
        if (is_ok((ss-i)->currentMove))
            (*(ss-i)->continuationHistory)[pc][to] << bonus;
  }


  // update_quiet_stats() updates move sorting heuristics

  void update_quiet_stats(const Position& pos, Stack* ss, Move move, int bonus) {

    if (ss->killers[0] != move)
    {
        ss->killers[1] = ss->killers[0];
        ss->killers[0] = move;
    }

    Color us = pos.side_to_move();
    Thread* thisThread = pos.this_thread();
    thisThread->mainHistory[us][from_to(move)] << bonus;
    update_continuation_histories(ss, pos.moved_piece(move), to_sq(move), bonus);

    if (type_of(pos.moved_piece(move)) != PAWN)
        thisThread->mainHistory[us][from_to(reverse_move(move))] << -bonus;

    if (is_ok((ss-1)->currentMove))
    {
        Square prevSq = to_sq((ss-1)->currentMove);
        thisThread->counterMoves[pos.piece_on(prevSq)][prevSq] = move;
    }
  }

  // When playing with strength handicap, choose best move among a set of RootMoves
  // using a statistical rule dependent on 'level'. Idea by Heinz van Saanen.

  Move Skill::pick_best(size_t multiPV) {

    const RootMoves& rootMoves = Threads.main()->rootMoves;
    static PRNG rng(now()); // PRNG sequence should be non-deterministic

    // RootMoves are already sorted by score in descending order
    Value topScore = rootMoves[0].score;
    int delta = std::min(topScore - rootMoves[multiPV - 1].score, PawnValueMg);
    int weakness = 120 - level;
    int maxScore = -VALUE_INFINITE;

    // Choose best move. For each move score we add two terms, both dependent on
    // weakness. One is deterministic and bigger for weaker levels, and one is
    // random. Then we choose the move with the resulting highest score.
    for (size_t i = 0; i < multiPV; ++i)
    {
        // This is our magic formula
        int push = (  weakness * int(topScore - rootMoves[i].score)
                    + delta * (rng.rand<unsigned>() % weakness)) / 128;

        if (rootMoves[i].score + push >= maxScore)
        {
            maxScore = rootMoves[i].score + push;
            best = rootMoves[i].pv[0];
        }
    }

    return best;
  }

} // namespace

/// MainThread::check_time() is used to print debug info and, more importantly,
/// to detect when we are out of available time and thus stop the search.

void MainThread::check_time() {

  if (--callsCnt > 0)
      return;

  // When using nodes, ensure checking rate is not lower than 0.1% of nodes
  callsCnt = Limits.nodes ? std::min(1024, int(Limits.nodes / 1024)) : 1024;

  static TimePoint lastInfoTime = now();

  TimePoint elapsed = Time.elapsed();
  TimePoint tick = Limits.startTime + elapsed;

  if (tick - lastInfoTime >= 1000)
  {
      lastInfoTime = tick;
      dbg_print();
  }

  // We should not stop pondering until told so by the GUI
  if (ponder)
      return;

  if (   (Limits.use_time_management() && (elapsed > Time.maximum() - 10 || stopOnPonderhit))
      || (Limits.movetime && elapsed >= Limits.movetime)
      || (Limits.nodes && Threads.nodes_searched() >= (uint64_t)Limits.nodes))
      Threads.stop = true;
}


/// UCI::pv() formats PV information according to the UCI protocol. UCI requires
/// that all (if any) unsearched PV lines are sent using a previous search score.

string UCI::pv(const Position& pos, Depth depth, Value alpha, Value beta) {

  std::stringstream ss;
  TimePoint elapsed = Time.elapsed() + 1;
  const RootMoves& rootMoves = pos.this_thread()->rootMoves;
  size_t pvIdx = pos.this_thread()->pvIdx;
  size_t multiPV = std::min((size_t)Options["MultiPV"], rootMoves.size());
  uint64_t nodesSearched = Threads.nodes_searched();
  uint64_t tbHits = Threads.tb_hits() + (TB::RootInTB ? rootMoves.size() : 0);

  for (size_t i = 0; i < multiPV; ++i)
  {
#if defined (Stockfish) || (Weakfish)
      bool updated = rootMoves[i].score != -VALUE_INFINITE;
#else
      bool updated = (i <= pvIdx && rootMoves[i].score != -VALUE_INFINITE);
#endif

      if (depth == 1 && !updated)
          continue;

      Depth d = updated ? depth : depth - 1;
      Value v = updated ? rootMoves[i].score : rootMoves[i].previousScore;
#ifdef Noir
      bool tb = TB::RootInTB && abs(v) < VALUE_TB_WIN - 5 * PawnValueEg;
#else
      bool tb = TB::RootInTB && abs(v) < VALUE_MATE - MAX_PLY;
#endif
      v = tb ? rootMoves[i].tbScore : v;

      if (ss.rdbuf()->in_avail()) // Not at first line
          ss << "\n";

      ss << "info"
         << " depth "    << d
         << " seldepth " << rootMoves[i].selDepth
         << " multipv "  << i + 1
         << " score "    << UCI::value(v);

      if (!tb && i == pvIdx)
          ss << (v >= beta ? " lowerbound" : v <= alpha ? " upperbound" : "");

      ss << " nodes "    << nodesSearched
         << " nps "      << nodesSearched * 1000 / elapsed;

      if (elapsed > 1000) // Earlier makes little sense
          ss << " hashfull " << TT.hashfull();

      ss << " tbhits "   << tbHits
         << " time "     << elapsed
         << " pv";

      for (Move m : rootMoves[i].pv)
          ss << " " << UCI::move(m, pos.is_chess960());
  }

  return ss.str();
}


/// RootMove::extract_ponder_from_tt() is called in case we have no ponder move
/// before exiting the search, for instance, in case we stop the search during a
/// fail high at root. We try hard to have a ponder move to return to the GUI,
/// otherwise in case of 'ponder on' we have nothing to think on.

bool RootMove::extract_ponder_from_tt(Position& pos) {

    StateInfo st;
    bool ttHit;

    assert(pv.size() == 1);

    if (pv[0] == MOVE_NONE)
        return false;

    pos.do_move(pv[0], st);
    TTEntry* tte = TT.probe(pos.key(), ttHit);

    if (ttHit)
    {
        Move m = tte->move(); // Local copy to be SMP safe
        if (MoveList<LEGAL>(pos).contains(m))
            pv.push_back(m);
    }

    pos.undo_move(pv[0]);
    return pv.size() > 1;
}

void Tablebases::rank_root_moves(Position& pos, Search::RootMoves& rootMoves) {

    RootInTB = false;
    UseRule50 = bool(Options["Syzygy50MoveRule"]);
    ProbeDepth = int(Options["SyzygyProbeDepth"]);
    Cardinality = int(Options["SyzygyProbeLimit"]);
#ifndef Noir
    bool dtz_available = true;
#endif
    // Tables with fewer pieces than SyzygyProbeLimit are searched with
    // ProbeDepth == DEPTH_ZERO
    if (Cardinality > MaxCardinality)
    {
        Cardinality = MaxCardinality;
        ProbeDepth = 0;
    }

    if (Cardinality >= popcount(pos.pieces()) && !pos.can_castle(ANY_CASTLING))
    {
        // Rank moves using DTZ tables
        RootInTB = root_probe(pos, rootMoves);

        if (!RootInTB)
        {
            // DTZ tables are missing; try to rank moves using WDL tables
#ifndef Noir
            dtz_available = false;
#endif
            RootInTB = root_probe_wdl(pos, rootMoves);
        }
    }

    if (RootInTB)
    {
        // Sort moves according to TB rank
        std::sort(rootMoves.begin(), rootMoves.end(),
                  [](const RootMove &a, const RootMove &b) { return a.tbRank > b.tbRank; } );

        // Probe during search only if DTZ is not available and we are winning
#ifndef Noir
        if (dtz_available || rootMoves[0].tbScore <= VALUE_DRAW)
            Cardinality = 0;
#endif
    }

    else
    {
        // Clean up if root_probe() and root_probe_wdl() have failed
        for (auto& m : rootMoves)
            m.tbRank = 0;
    }
}
